package ah2912;
import robocode.*;
import java.awt.Color;

/**
 * SnippetBot - a robot by Alisdair Owens
 * This bot includes all sorts of useful snippets.  It is not
 * designed to be a good fighter (although it does well 1v1),
 * just to show how certain things are done
 * Bits of code lifted from Nicator and Chrisbot
 * Conventions in this bot include: Use of radians throughout
 * Storing absolute positions of enemy bots rather than relative ones
 * Very little code in events
 * These are all good programming practices for robocode
 * There may also be methods that arent used; these might just be useful for you.
 */
public class Ciclone extends AdvancedRobot
{
	/**
	 * run: SnippetBot's default behavior
	 */
	Enemy target;					//our current enemy
	final double PI = Math.PI;			//just a constant
	int direction = 1;				//direction we are heading...1 = forward, -1 = backwards
	double firePower;       			//the power of the shot we will be using - set by do firePower()
	boolean first = true;
	boolean peek;		// Don't turn if there's a robot there
	double moveAmount;	// How much to move
	double x,y,xo,yo;

	public void run() {
		target = new Enemy();
		target.distance = 100000;			//initialise the distance so that we can select a target
		setColors(Color.white,Color.red,Color.white);	//sets the colours of the robot

		//the next two lines mean that the turns of the robot, gun and radar are independant
		setAdjustGunForRobotTurn(true);
		setAdjustRadarForGunTurn(true);
		turnRadarRightRadians(2*PI);			//turns the radar right around to get a view of the field
		xo = -1;
		yo = -1;
		while(true) {
			if(getEnergy()>60){
				doMovement();				//Move the bot
				doFirePower();				//select the fire power to use
				doScanner();				//Oscillate the scanner over the bot
				doGun();
				out.println(target.distance);		//move the gun to predict where the enemy will be
				fire(firePower);
				execute();				//execute all commands
			}
			else{
				if(first){
					// Initialize moveAmount to the maximum possible for this battlefield.
					moveAmount = Math.max(getBattleFieldWidth(),getBattleFieldHeight());
					// Initialize peek to false
					peek = false;
			
					// turnLeft to face a wall.
					// getHeading() % 90 means the remainder of 
					// getHeading() divided by 90.
					turnLeft(getHeading() % 90);
					ahead(moveAmount);
					// Turn the gun to turn right 90 degrees.
					peek = true;
					turnGunRight(90);
					turnRight(90);
		
					first = false;
				}
				// Look before we turn when ahead() completes.
				peek = true;
				// Move up the wall
				ahead(moveAmount);
				// Don't look now
				peek = false;
				// Turn to the next wall
				turnRight(90);
			}				
		}
	}

	void doMovement() {
		if (getTime()%20 == 0)  { 		//every twenty 'ticks'
			direction = 1 - direction;		//reverse direction
			x=getX();
			y=getY();
			if(x==xo && y==yo){
				direction = 1 - direction;		//reverse direction
			}
			xo=x;
			yo=y;
			setAhead(direction*300);	//move in that direction
		}
		setTurnRightRadians(target.bearing + (PI/2)); //every turn move to circle strafe the enemy
	}
	
	void doFirePower() {
		if(target.distance<100){
			if(target.energy<4){
				firePower = target.energy / 4;
			}
			else {
				firePower = (target.energy + 2) / 6;
			}
			if(firePower>3){
				firePower = 3;
			}
		}
		else{
			if(target.energy<4){
				firePower = target.energy / 4;
			}
			else {
				firePower = (target.energy + 2) / 6;
			}
			if(firePower>1){
				firePower = 1;
			}
		}
	}

	void doScanner() {
		double radarOffset;
		if (getTime() - target.ctime > 4) { 	//if we haven't seen anybody for a bit....
			radarOffset = 360;		//rotate the radar to find a target
		} else {	
			
			//next is the amount we need to rotate the radar by to scan where the target is now
			radarOffset = getRadarHeadingRadians() - absbearing(getX(),getY(),target.x,target.y);
			
			//this adds or subtracts small amounts from the bearing for the radar to produce the wobbling
			//and make sure we don't lose the target
			if (radarOffset < 0)
				radarOffset -= PI/8;
			else
				radarOffset += PI/8; 
		}
		//turn the radar
		setTurnRadarLeftRadians(NormaliseBearing(radarOffset));
	}
	
	void doGun() {
		
		//works out how long it would take a bullet to travel to where the enemy is *now*
		//this is the best estimation we have
		long time = getTime() + (int)(target.distance/(20-(3*firePower)));
		
		//offsets the gun by the angle to the next shot based on linear targeting provided by the enemy class
		double gunOffset = getGunHeadingRadians() - absbearing(getX(),getY(),target.guessX(time),target.guessY(time));
		setTurnGunLeftRadians(NormaliseBearing(gunOffset));
	}
	
	
	//if a bearing is not within the -pi to pi range, alters it to provide the shortest angle
	double NormaliseBearing(double ang) {
		if (ang > PI)
			ang -= 2*PI;
		if (ang < -PI)
			ang += 2*PI;
		return ang;
	}
	
	//if a heading is not within the 0 to 2pi range, alters it to provide the shortest angle
	double NormaliseHeading(double ang) {
		if (ang > 2*PI)
			ang -= 2*PI;
		if (ang < 0)
			ang += 2*PI;
		return ang;
	}
	
	//returns the distance between two x,y coordinates
	public double getrange( double x1,double y1, double x2,double y2 )
	{
		double xo = x2-x1;
		double yo = y2-y1;
		double h = Math.sqrt( xo*xo + yo*yo );
		return h;	
	}
	
	//gets the absolute bearing between to x,y coordinates
	public double absbearing( double x1,double y1, double x2,double y2 )
	{
		double xo = x2-x1;
		double yo = y2-y1;
		double h = getrange( x1,y1, x2,y2 );
		if( xo > 0 && yo > 0 )
		{
			return Math.asin( xo / h );
		}
		if( xo > 0 && yo < 0 )
		{
			return Math.PI - Math.asin( xo / h );
		}
		if( xo < 0 && yo < 0 )
		{
			return Math.PI + Math.asin( -xo / h );
		}
		if( xo < 0 && yo > 0 )
		{
			return 2.0*Math.PI - Math.asin( -xo / h );
		}
		return 0;
	}


	/**
	 * onScannedRobot: What to do when you see another robot
	 */
	public void onScannedRobot(ScannedRobotEvent e) {
		if(getEnergy()>60){
			//if we have found a closer robot....
			if ((e.getDistance() < target.distance)||(target.name == e.getName())) {
				//the next line gets the absolute bearing to the point where the bot is
				double absbearing_rad = (getHeadingRadians()+e.getBearingRadians())%(2*PI);
				//this section sets all the information about our target
				target.name = e.getName();
				target.x = getX()+Math.sin(absbearing_rad)*e.getDistance(); //works out the x coordinate of where the target is
				target.y = getY()+Math.cos(absbearing_rad)*e.getDistance(); //works out the y coordinate of where the target is
				target.bearing = e.getBearingRadians();
				target.head = e.getHeadingRadians();
				target.ctime = getTime();				//game time at which this scan was produced
				target.speed = e.getVelocity();
				target.distance = e.getDistance();
				target.energy = e.getEnergy();
			}
		}
		else{
			fire(2);
			// Note that scan is called automatically when the robot is moving.
			// By calling it manually here, we make sure we generate another scan event if there's a robot on the next 
			// wall, so that we do not start moving up it until it's gone.
			if (peek)
				scan();
		}
	}

	public void onRobotDeath(RobotDeathEvent e) {
		if(getEnergy()<=60){
			if (e.getName() == target.name)
				target.distance = 10000; //this will effectively make it search for a new target
		}
	}
	
	/**
	 * onHitRobot:  Move away a bit.
	 */	
	public void onHitRobot(HitRobotEvent e) {
		if(getEnergy()>60){
			// If he's in front of us, set back up a bit.
			if (e.getBearing() > -90 && e.getBearing() < 90)
				back(100);
			// else he's in back of us, so set ahead a bit.
			else
				ahead(100);
		}
	}

}

class Enemy {
	/*
	 * ok, we should really be using accessors and mutators here,
	 * (i.e getName() and setName()) but life's too short.
	 */
	String name;
	public double bearing;
	public double head;
	public long ctime; //game time that the scan was produced
	public double speed;
	public double x,y;
	public double distance;
	public double energy;
	public double guessX(long when)
	{
		long diff = when - ctime;
		return x+Math.sin(head)*speed*diff;
	}
	public double guessY(long when)
	{
		long diff = when - ctime;
		return y+Math.cos(head)*speed*diff;
	}

}
