package cslib;
import robocode.*;
import java.awt.Color;


public class faiscas extends AdvancedRobot
{
        
        Enemy target;					//our current enemy
        final double PI = Math.PI;			//just a constant
        int direction = 1;				//direction we are heading...1 = forward, -1 = backwards
        double firePower;       			//the power of the shot we will be using - set by do firePower()

        public void run() {
                target = new Enemy();
                target.distance = 100000;			//initialise the distance so that we can select a target
                setColors(Color.red,Color.blue,Color.green);	//sets the colours of the robot

                //the next two lines mean that the turns of the robot, gun and radar are independant
                setAdjustGunForRobotTurn(true);
                setAdjustRadarForGunTurn(true);
                turnRadarRightRadians(2*PI);			//turns the radar right around to get a view of the field
                while(true) {
                        doMovement();				//Move the bot
                        doFirePower();				//select the fire power to use
                        doScanner();				//Oscillate the scanner over the bot
                        doGun();
                        out.println(target.distance);		//move the gun to predict where the enemy will be
                        fire(firePower);
                        execute();				//execute all commands
                }
        }


        void doFirePower() {
                firePower = 400/target.distance;//selects a bullet power based on our distance away from the target
        }

        void doMovement() {
                if (getTime()%20 == 0)  { 		//every twenty 'ticks'
                        direction *= -1;
                        turnRight(30);		
                        setAhead(direction*300);	//move in that direction
                }
                setTurnRightRadians(target.bearing + (PI/2)); //every turn move to circle strafe the enemy
        }

        void doScanner() {
                double radarOffset;
                if (getTime() - target.ctime > 4) { 	
                        radarOffset = 360;		
                } else {

                        //next is the amount we need to rotate the radar by to scan where the target is now
                        radarOffset = getRadarHeadingRadians() - absbearing(getX(),getY(),target.x,target.y);

                        //this adds or subtracts small amounts from the bearing for the radar to produce the wobbling
                        //and make sure we don't lose the target
                        if (radarOffset < 0)
                        radarOffset -= PI/8;
                        else
                        radarOffset += PI/8;
                }
                //turn the radar
                setTurnRadarLeftRadians(NormaliseBearing(radarOffset));
        }


	public void onHitWallEvent(HitWallEvent e)
	{
		turnRight(e.getBearing()+180);
		ahead(50);
	}
	
	public void OnHitRobotEvent (HitRobotEvent e)
	{
		 turnGunRight(getGunHeading()-e.getBearing());
		 fire(3);
		 turnRight(45);
		 ahead(100);
	}
	
	public void OnHitByBulletEvent(HitByBulletEvent e)
	{
		
		turnGunRight(getGunHeading() - e.getBearing());
	    fire(3);
	    turnRight(45);
	ahead(100);
		
	}
	
        void doGun() {

                
                long time = getTime() + (int)(target.distance/(20-(3*firePower)));

                //offsets the gun by the angle to the next shot based on linear targeting provided by the enemy class
                double gunOffset = getGunHeadingRadians() - absbearing(getX(),getY(),target.guessX(time),target.guessY(time));
                setTurnGunLeftRadians(NormaliseBearing(gunOffset));
        }



        double NormaliseBearing(double ang) {
                if (ang > PI)
                ang -= 2*PI;
                if (ang < -PI)
                ang += 2*PI;
                return ang;
        }


        double NormaliseHeading(double ang) {
                if (ang > 2*PI)
                ang -= 2*PI;
                if (ang < 0)
                ang += 2*PI;
                return ang;
        }


        public double getrange( double x1,double y1, double x2,double y2 )
        {
                double xo = x2-x1;
                double yo = y2-y1;
                double h = Math.sqrt( xo*xo + yo*yo );
                return h;
        }


        public double absbearing( double x1,double y1, double x2,double y2 )
        {
                double xo = x2-x1;
                double yo = y2-y1;
                double h = getrange( x1,y1, x2,y2 );
                if( xo > 0 && yo > 0 )
                {
                        return Math.asin( xo / h );
                }
                if( xo > 0 && yo < 0 )
                {
                        return Math.PI - Math.asin( xo / h );
                }
                if( xo < 0 && yo < 0 )
                {
                        return Math.PI + Math.asin( -xo / h );
                }
                if( xo < 0 && yo > 0 )
                {
                        return 2.0*Math.PI - Math.asin( -xo / h );
                }
                return 0;
        }


        /**
        * onScannedRobot: What to do when you see another robot
        */
        public void onScannedRobot(ScannedRobotEvent e) {

                if ((e.getDistance() < target.distance)||(target.name == e.getName())) {

                        double absbearing_rad = (getHeadingRadians()+e.getBearingRadians())%(2*PI);

                        target.name = e.getName();
                        target.x = getX()+Math.sin(absbearing_rad)*e.getDistance(); 
                        target.y = getY()+Math.cos(absbearing_rad)*e.getDistance(); 
                        target.bearing = e.getBearingRadians();
                        target.head = e.getHeadingRadians();
                        target.ctime = getTime();				
                        target.speed = e.getVelocity();
                        target.distance = e.getDistance();
                }
        }

        public void onRobotDeath(RobotDeathEvent e) {
                if (e.getName() == target.name)
                target.distance = 10000; //this will effectively make it search for a new target
        }

}

class Enemy {
        /*
        * ok, we should really be using accessors and mutators here,
        * (i.e getName() and setName()) but life's too short.
        */
        String name;
        public double bearing;
        public double head;
        public long ctime; //game time that the scan was produced
        public double speed;
        public double x,y;
        public double distance;
        public double guessX(long when)
        {
                long diff = when - ctime;
                return x+Math.sin(head)*speed*diff;
        }
        public double guessY(long when)
        {
                long diff = when - ctime;
                return y+Math.cos(head)*speed*diff;
        }

}